#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QElapsedTimer>
#include <QDebug>
#include <QMessageBox>
#include <QException>

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    ui->glWidget->installEventFilter(this);
}

bool MainWindow::Init(QSplashScreen &splash)
{
    int nAlignment = Qt::AlignLeft | Qt::AlignBottom;

    // openal
    splash.showMessage("Initialize OpenAL...", nAlignment);
    splash.update();
    alutInit(&__argc, __argv);

    splash.showMessage("Initialize OpenGL...", nAlignment);
    splash.update();
    if (false == InitGL(ui->glWidget))
    {
        QMessageBox msgBox;
        msgBox.setWindowTitle("ERROR!");
        msgBox.setText("OpenGL 4.0 not supported!");
        msgBox.setStandardButtons(QMessageBox::Ok);
        msgBox.exec();

        return false;
    }

    splash.showMessage("Initilaize BulletPhysics...", nAlignment);
    splash.update();
    if (false == InitPhysics())
    {
        QMessageBox msgBox;
        msgBox.setWindowTitle("ERROR!");
        msgBox.setText("BulletPhysics: init error!");
        msgBox.setStandardButtons(QMessageBox::Ok);
        msgBox.exec();

        return false;
    }

    splash.showMessage("Loading Scene...", nAlignment);
    splash.update();
    if (false == InitScene())
    {
        return false;
    }

    splash.clearMessage();
    splash.update();

    showMaximized();
    QApplication::setOverrideCursor(Qt::BlankCursor);

    m_elapsedTimer.start();
    m_nElapsedTime = m_nCurrentTime = m_elapsedTimer.nsecsElapsed();

    connect(&m_Timer, SIGNAL(timeout()), this, SLOT(TimerTick()));
    m_Timer.start(0);

    return true;
}

MainWindow::~MainWindow()
{
    m_Timer.stop();
    disconnect(&m_Timer, SIGNAL(timeout()), this, SLOT(TimerTick()));

    QApplication::setOverrideCursor(Qt::ArrowCursor);

    m_SkyBox.Release();
    m_modelDraw.Release();
    m_modelPhysics.Release();
    ExitPhysics();
    delete ui;
}

bool MainWindow::InitGL(QWidget *pWidget)
{
    hWnd = (HWND)pWidget->winId();
    hDC = GetDC(hWnd);

    PIXELFORMATDESCRIPTOR pfd =
    {
        sizeof(PIXELFORMATDESCRIPTOR),                  // Size Of This Pixel Format Descriptor
        1,                              // Version Number
        PFD_DRAW_TO_WINDOW |                        // Format Must Support Window
        PFD_SUPPORT_OPENGL |                        // Format Must Support OpenGL
        PFD_DOUBLEBUFFER,                       // Must Support Double Buffering
        PFD_TYPE_RGBA,                          // Request An RGBA Format
        32,                               // Select Our Color Depth
        0, 0, 0, 0, 0, 0,                       // Color Bits Ignored
        0,                              // No Alpha Buffer
        0,                              // Shift Bit Ignored
        0,                              // No Accumulation Buffer
        0, 0, 0, 0,                         // Accumulation Bits Ignored
        32,                             // Z-Buffer (Depth Buffer)
        0,                              // Stencil Buffer
        0,                              // No Auxiliary Buffer
        PFD_MAIN_PLANE,                         // Main Drawing Layer
        0,                              // Reserved
        0, 0, 0                             // Layer Masks Ignored
    };

    int pf = ChoosePixelFormat(hDC, &pfd);
    SetPixelFormat(hDC, pf, &pfd);
    hRC = wglCreateContext(hDC);
    wglMakeCurrent(hDC, hRC);

    glewInit();

    if (false == glewIsSupported("GL_VERSION_4_0"))
    {
        return false;
    }

    return true;
}

bool MainWindow::InitScene()
{
    wglMakeCurrent(hDC, hRC);

    // opengl
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
    //glEnable(GL_BLEND);
    //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    wglSwapIntervalEXT(0);

    m_shaderDraw.Load("Shaders/Draw.vs.txt", "Shaders/Draw.fs.txt");
    m_shaderShadowMap.Load("Shaders/DrawToDepthTexture.vs.txt", "Shaders/DrawToDepthTexture.fs.txt");
    std::vector<int> listTypes;
    listTypes.push_back(GL_RGBA32F);
    m_RenderToShadowTexture.Load(listTypes, nShadowWidth, nShadowWidth);
    m_SkyBox.Load("Scene/", "barren_bk.jpg", "barren_dn.jpg", "barren_ft.jpg", "barren_lf.jpg", "barren_rt.jpg", "barren_up.jpg");

    // Models
    // 1/2 - draw
    m_modelDraw.Load("Scene", "Scene.obj");
    m_modelDraw.CreateOpenGLBuffers();
    // 2/2 - physics
    m_modelPhysics.Load("Scene", "Physics.obj", glm::mat4(1.0f), false);
    CreateConcaveMesh(glm::vec3(0,0,0), glm::vec3(0,0,0), 0.0f, 0.0f, 1.0f, m_modelPhysics.GetVertices());

    // avatar
    rigidbodyAvatar = CreateAvatar(glm::vec3(15, 3, 15), m_fRadius, m_fHeight);
    m_Camera.Init(glm::vec3(0,8,20), glm::vec3(0,0,0));

    // box
    m_modelBox.Load("Scene", "barrel.obj", glm::scale(glm::vec3(0.015f, 0.015f, 0.015f)), false);
    m_modelBox.CreateOpenGLBuffers();

    for(int i = 0; i < numTextures; i++)
    {
        QString strFilename = "Scene/" + strFilenames[i];

        Texture texture;
        texture.Load(strFilename.toStdString());
        textures[i] = texture.ID();
    }

    srand(time(NULL));
    for(int x = -2; x <= 2; x++)
    {
        for(int z = -2; z <= 2; z++)
        {
            for(int y = 0; y < (4/*100db*/ * 6/*600db*/); y++)
            {
                float fScale = 2.1f;
                btRigidBody *body = CreateConvexMesh(glm::vec3(x * fScale, 20 + (y * fScale), z * fScale), glm::vec3(0,0,0), 120.0f, 0.05f, 1.0f, m_modelBox.GetVertices());
                m_listRigidBodies.push_back(body);

                int id = rand() % numTextures;
                m_listRigidBodiesTextureId.push_back( textures[id] );
            }
        }
    }

    // openal
    ALuint buffer;
    buffer = alutCreateBufferFromFile("Background.wav");
    ALuint source;
    alGenSources(1, &source);
    alSourcei(source, AL_BUFFER, buffer);
    alSourcei(source, AL_LOOPING, true);
    alSourcePlay(source);

    return true;
}

btRigidBody* MainWindow::CreateConvexMesh(glm::vec3 v3Translate, glm::vec3 v3RotateEuler, float fMass, float fRestitution, float fFriction, std::vector<Vertex> *pListVertices)
{
    btConvexHullShape* convexHullShape = new btConvexHullShape();
    convexHullShape->setLocalScaling(btVector3(1.0f, 1.0f, 1.0f));

    for(int i = 0; i < pListVertices->size(); i++)
    {
        glm::vec3 v = pListVertices->at(i).v3Position;
        convexHullShape->addPoint(btVector3(v.x, v.y, v.z));
    }

    convexHullShape->setMargin(0.001f);

    btTransform trans;
    trans.setIdentity();
    btQuaternion quat(v3RotateEuler.x, v3RotateEuler.y, v3RotateEuler.z);
    trans.setRotation(quat);
    trans.setOrigin(btVector3(v3Translate.x, v3Translate.y, v3Translate.z));

    btVector3 localInertia(0, 0, 0);
    convexHullShape->calculateLocalInertia(fMass, localInertia);

    btDefaultMotionState* myMotionState = new btDefaultMotionState(trans);
    btRigidBody::btRigidBodyConstructionInfo rbInfo(fMass, myMotionState, convexHullShape, localInertia);
    rbInfo.m_restitution = fRestitution;
    rbInfo.m_friction = fFriction;
    rbInfo.m_linearDamping = 0.1f;
    rbInfo.m_angularDamping = 0.1f;

    btRigidBody* body = new btRigidBody(rbInfo);

    dynamicsWorld->addRigidBody(body);

    return body;
}

btRigidBody* MainWindow::CreateConcaveMesh(glm::vec3 v3Translate, glm::vec3 v3RotateEuler, float fMass, float fRestitution, float fFriction, std::vector<Vertex> *pListVertices)
{
    glm::mat4 mRotateX = glm::rotate(glm::mat4(1.0f), v3RotateEuler.x, glm::vec3(1,0,0));
    glm::mat4 mRotateY = glm::rotate(glm::mat4(1.0f), v3RotateEuler.y, glm::vec3(0,1,0));
    glm::mat4 mRotateZ = glm::rotate(glm::mat4(1.0f), v3RotateEuler.z, glm::vec3(0,0,1));
    glm::mat4 mTramslate = glm::translate(glm::mat4(1.0f), v3Translate);
    glm::mat4 mTransform = mTramslate * mRotateZ * mRotateY * mRotateX;

    btVector3 *btVerts = new btVector3[pListVertices->size()]();
    int* btInd = new int[pListVertices->size()];

    for(int i = 0; i < (int)pListVertices->size(); i++)
    {
        glm::vec3 vIn = pListVertices->at(i).v3Position;
        glm::vec3 vOut = glm::vec3(mTransform * glm::vec4(vIn.x, vIn.y, vIn.z, 1.0f));

        btVerts[i] = btVector3(vOut.x, vOut.y, vOut.z);
        btInd[i] = i;
    }

    int totalVerts = pListVertices->size();
    int totalTriangles = pListVertices->size() / 3;
    int vertStride = sizeof(btVector3);
    int indexStride = 3 * sizeof(int);

    btTriangleIndexVertexArray* indexVertexArrays = new btTriangleIndexVertexArray(
            totalTriangles,
            btInd,
            indexStride,
            totalVerts,
            (btScalar*) &btVerts[0].x(),
            vertStride
            );

    bool useQuantizedAabbCompression = true;
    btBvhTriangleMeshShape* trimeshShape = new btBvhTriangleMeshShape(indexVertexArrays, useQuantizedAabbCompression);
    trimeshShape->setMargin(0.001f);

    btVector3 localInertia(0, 0, 0);
    if (fMass > 0.0f)
    {
        trimeshShape->calculateLocalInertia(fMass, localInertia);
    }
    btDefaultMotionState* myMotionState = new btDefaultMotionState();
    btRigidBody::btRigidBodyConstructionInfo cInfo( fMass, myMotionState, trimeshShape, localInertia);
    cInfo.m_restitution = fRestitution;
    cInfo.m_friction = fFriction;

    btRigidBody *rigidBody = new btRigidBody(cInfo);
    dynamicsWorld->addRigidBody(rigidBody);

    return rigidBody;
}

btRigidBody* MainWindow::CreateAvatar(glm::vec3 v3Pos, float radius, float height)
{
    btCollisionShape* colShape = new btCapsuleShape(radius, height);
    btVector3 localInertia(0, 0, 0);
    colShape->calculateLocalInertia(85.0f, localInertia);
    colShape->setMargin(0.001f);

    btTransform startTransform;
    startTransform.setIdentity();
    startTransform.setOrigin(btVector3(v3Pos.x, v3Pos.y, v3Pos.z));
    btQuaternion orn; orn.setEuler(0, 0, 0);
    startTransform.setRotation(orn);

    btDefaultMotionState* myMotionState = new btDefaultMotionState(startTransform);
    btRigidBody::btRigidBodyConstructionInfo rbInfo(85.0f, myMotionState, colShape, localInertia);
    rbInfo.m_restitution = 0.0f;
    rbInfo.m_friction = 0.3f;

    btRigidBody *rigidBody = new btRigidBody(rbInfo);
    dynamicsWorld->addRigidBody(rigidBody);

    return rigidBody;
}

bool MainWindow::InitPhysics()
{
    btSetTaskScheduler( btCreateDefaultTaskScheduler() );
    broadphase = new btDbvtBroadphase();
    collisionConfiguration = new btDefaultCollisionConfiguration();
    dispatcher = new btCollisionDispatcherMt(collisionConfiguration, 40);
    solverPool = new btConstraintSolverPoolMt(BT_MAX_THREAD_COUNT);
    parallelSolver = new btSequentialImpulseConstraintSolverMt();
    dynamicsWorld = new btDiscreteDynamicsWorldMt(dispatcher, broadphase, solverPool, parallelSolver, collisionConfiguration);

    dynamicsWorld->setGravity(btVector3(0, -9.81f, 0));

    return true;
}

void MainWindow::ExitPhysics()
{
    //delete dynamicsWorld;
    //delete solver;
    //delete overlappingPairCache;
    //delete dispatcher;
    //delete collisionConfiguration;
}

void MainWindow::TimerTick()
{
    // Update
    wglMakeCurrent(hDC, hRC);

    // getDT
    m_nElapsedTime = m_nCurrentTime;
    m_nCurrentTime = m_elapsedTimer.nsecsElapsed();
    dt = (float)(m_nCurrentTime - m_nElapsedTime) / 1000000000.0f;

    // print fps
    nFPS++;
    fSec += dt;
    if (fSec >= 1.0f)
    {
        this->setWindowTitle("FPS: " + QString::number(nFPS));

        nFPS = 0;
        fSec = 0.0f;
    }

    if (dt > 1.0f / 20.0f)
    {
        dt = 1.0f / 20.0f;
    }
    if (dt <= 0.0f)
    {
        dt = 1.0f / 60.0f;
    }

    int nWidth = ui->glWidget->width();
    int nHeight = ui->glWidget->height();
    if (nWidth < 1) { nWidth = 1; }
    if (nHeight < 1) { nHeight = 1; }

    // physics
    dynamicsWorld->stepSimulation(dt, 1);

    // mouse rotate
    QPoint pointDiff = cursor().pos() - QPoint(width() / 2, height() / 2);
    cursor().setPos(QPoint(width() / 2, height() / 2));
    m_Camera.Rotate(pointDiff.x(), pointDiff.y());
    m_Camera.Update(dt);

    rigidbodyAvatar->activate(true); // not sleeping
    rigidbodyAvatar->setAngularFactor(btVector3(0,0,0)); // not rotate

    btVector3 pos = rigidbodyAvatar->getCenterOfMassPosition();
    m_Camera.SetPos(glm::vec3(pos.x(), pos.y() + (m_fHeight / 2.0f) + (m_fRadius / 2.0f), pos.z()));
    glm::vec3 v3CameraPos = m_Camera.GetPos();
    glm::vec3 v3CameraAt = m_Camera.GetAt();
    glm::vec3 v3CameraDir = glm::normalize(v3CameraAt - v3CameraPos);

    float speed = 4.0f;
    glm::vec3 dir(v3CameraDir.x, 0, v3CameraDir.z);
    if (glm::length(dir) > 0.0001f)
    {
        dir = glm::normalize(dir);
    }
    else
    {
        dir = glm::vec3(0, 0, 0);
    }

    btVector3 btLinVel = rigidbodyAvatar->getLinearVelocity();
    glm::vec3 v3LinearVelocity(btLinVel.x(), btLinVel.y(), btLinVel.z());

    // move
    glm::vec3 vel(0, 0, 0);
    if (true == m_bKeys[Qt::Key_W])
    {
        vel += dir;
    }
    if (true == m_bKeys[Qt::Key_S])
    {
        vel += -dir;
    }
    if (true == m_bKeys[Qt::Key_D])
    {
        vel += glm::cross(dir, glm::vec3(0, 1, 0));
    }
    if (true == m_bKeys[Qt::Key_A])
    {
        vel += -glm::cross(dir, glm::vec3(0, 1, 0));
    }

    if (glm::length(vel) > 0.0001f)
    {
        vel = glm::normalize(vel);
        vel *= speed;
    }

    v3LinearVelocity = glm::vec3(vel.x, v3LinearVelocity.y, vel.z);
    rigidbodyAvatar->setLinearVelocity(btVector3(v3LinearVelocity.x, v3LinearVelocity.y, v3LinearVelocity.z));

    glm::vec3 v3LightPos = glm::vec3(32.6785f, 85.7038f, -39.8369f);
    glm::vec3 v3LightAt = glm::vec3(0, 0, 0);
    glm::vec3 v3LightDir = glm::normalize(v3LightAt - v3LightPos);

    //glm::mat4 mLightWorld = glm::mat4(1.0f);
    glm::mat4 mLightView = glm::lookAtRH(v3LightPos, v3LightAt, glm::vec3(0, 1, 0));
    glm::mat4 mLightProj = glm::orthoRH(-30.0f,30.0f, -30.0f,30.0f, 1.0f, 200.0f);



    // draw
    // 1/2 - draw to shadow texture
    m_RenderToShadowTexture.Bind();

    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    glViewport(0, 0, nShadowWidth, nShadowWidth);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    m_shaderShadowMap.Begin();
    glm::mat4 mWorld = glm::mat4(1.0f);
    m_shaderShadowMap.SetMatrix("matWorld", &mWorld);
    m_shaderShadowMap.SetMatrix("matView", &mLightView);
    m_shaderShadowMap.SetMatrix("matProj", &mLightProj);

    m_modelDraw.Draw(&m_shaderShadowMap);

    for(int i = 0; i < m_listRigidBodies.count(); i++)
    {
        btRigidBody *body = m_listRigidBodies.at(i);

        btTransform trans = body->getWorldTransform();
        btQuaternion quat = trans.getRotation();
        float fAngle = quat.getAngle();
        btVector3 axis = quat.getAxis();
        btVector3 tr = trans.getOrigin();

        glm::mat4 mWorld = glm::translate(glm::vec3(tr.x(), tr.y(), tr.z()) ) * glm::rotate(fAngle, glm::vec3(axis.x(), axis.y(), axis.z()));

        m_shaderShadowMap.SetMatrix("matWorld", &mWorld);

        m_shaderShadowMap.SetTexture("g_Texture", m_listRigidBodiesTextureId.at(i), 0);
        m_modelBox.Draw(&m_shaderShadowMap);
    }

    m_shaderShadowMap.End();
    m_RenderToShadowTexture.Unbind();

    // 2/2 - draw to screen with shadow
    glm::mat4 mCameraView = glm::lookAtRH(v3CameraPos, v3CameraAt, glm::vec3(0, 1, 0));
    glm::mat4 mCameraProj = glm::perspectiveRH(glm::radians(45.0f), (float)nWidth / (float)nHeight, 0.1f, 1000.0f);

    glClearColor(0.5f, 0.5f, 1.0f, 1.0f);
    glViewport(0, 0, nWidth, nHeight);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    m_shaderDraw.Begin();
    mWorld = glm::mat4(1.0f);
    m_shaderDraw.SetMatrix("matWorld", &mWorld);
    m_shaderDraw.SetMatrix("matView", &mCameraView);
    m_shaderDraw.SetMatrix("matProj", &mCameraProj);
    m_shaderDraw.SetMatrix("matLightView", &mLightView);
    m_shaderDraw.SetMatrix("matLightProj", &mLightProj);
    m_shaderDraw.SetVector3("lightDir", &v3LightDir);
    m_shaderDraw.SetTexture("g_DepthTexture", m_RenderToShadowTexture.GetTextureID(0), 1);

    m_modelDraw.Draw(&m_shaderDraw);

    for(int i = 0; i < m_listRigidBodies.count(); i++)
    {
        btRigidBody *body = m_listRigidBodies.at(i);

        btTransform trans = body->getWorldTransform();
        btQuaternion quat = trans.getRotation();
        float fAngle = quat.getAngle();
        btVector3 axis = quat.getAxis();
        btVector3 tr = trans.getOrigin();

        glm::mat4 mWorld = glm::translate(glm::vec3(tr.x(), tr.y(), tr.z()) ) * glm::rotate(fAngle, glm::vec3(axis.x(), axis.y(), axis.z()));

        m_shaderDraw.SetMatrix("matWorld", &mWorld);

        m_shaderDraw.SetTexture("g_Texture", m_listRigidBodiesTextureId.at(i), 0);
        m_modelBox.Draw(&m_shaderDraw);
    }

    m_shaderDraw.DisableTexture(1);
    m_shaderDraw.End();

    // sky
    glMatrixMode(GL_PROJECTION);
    glLoadMatrixf(glm::value_ptr(mCameraProj));
    glMatrixMode(GL_MODELVIEW);
    mWorld = glm::mat4(1.0f);
    glLoadMatrixf(glm::value_ptr(mCameraView * mWorld));

    m_SkyBox.Draw(v3CameraPos, 300.0f);

    SwapBuffers(hDC);
}

bool MainWindow::eventFilter(QObject *obj, QEvent *event)
{
    bool ret = QObject::eventFilter(obj, event);

    if (event->type() == QEvent::MouseMove)
    {
        //int nWidth = ui->glWidget->width();
        //int nHeight = ui->glWidget->height();
    }

    return ret;
}

void MainWindow::keyPressEvent(QKeyEvent *event)
{
    int id = event->key();
    if (id >= 0 && id <= 255)
    {
        m_bKeys[id] = true;
    }

    if (Qt::Key_Escape == event->key())
    {
        close();
    }
}

void MainWindow::keyReleaseEvent(QKeyEvent *event)
{
    int id = event->key();
    if (id >= 0 && id <= 255)
    {
        m_bKeys[id] = false;
    }
}

void MainWindow::mousePressEvent(QMouseEvent *event)
{
    //m_bMouseButtonDown = true;
}

void MainWindow::mouseReleaseEvent(QMouseEvent *event)
{
    //m_bMouseButtonDown = false;
}

void MainWindow::mouseMoveEvent(QMouseEvent *event)
{

}
